@@ -32,7 +32,7 @@
 static int nibblemap[] = {4, 3, 3, 2, 3, 2, 2, 1, 3, 2, 2, 1, 2, 1, 1, 0};
 
 static inline
-unsigned long count_free (char * map, unsigned int numchars)
+unsigned long count_free (unsigned char *map, unsigned int numchars)
 {
 	unsigned int i;
 	unsigned long sum = 0;
