@@ -79,6 +79,9 @@ do { pthread_spin_lock (&ext2s_pager_stats.lock);			      \
 #else /* !STATS */
 #define STAT_INC(field) /* nop */0
 #endif /* STATS */
+
+static void
+disk_cache_info_free_push (struct disk_cache_info *p);
 
 #define FREE_PAGE_BUFS 24
 
@@ -433,7 +436,7 @@ disk_pager_read_page (vm_offset_t page, void **buf, int *writelock)
     + offset % block_size;
   disk_cache_info[index].flags |= DC_INCORE;
   disk_cache_info[index].flags &=~ DC_UNTOUCHED;
-#ifndef NDEBUG
+#ifdef DEBUG_DISK_CACHE
   disk_cache_info[index].last_read = disk_cache_info[index].block;
   disk_cache_info[index].last_read_xor
     = disk_cache_info[index].block ^ DISK_CACHE_LAST_READ_XOR;
@@ -469,7 +472,7 @@ disk_pager_write_page (vm_offset_t page, void *buf)
   assert (disk_cache_info[index].block != DC_NO_BLOCK);
   offset = ((store_offset_t) disk_cache_info[index].block << log2_block_size)
     + offset % block_size;
-#ifndef NDEBUG			/* Not strictly needed.  */
+#ifdef DEBUG_DISK_CACHE			/* Not strictly needed.  */
   assert ((disk_cache_info[index].last_read ^ DISK_CACHE_LAST_READ_XOR)
 	  == disk_cache_info[index].last_read_xor);
   assert (disk_cache_info[index].last_read
@@ -538,6 +541,9 @@ disk_pager_notify_evict (vm_offset_t page)
 
   pthread_mutex_lock (&disk_cache_lock);
   disk_cache_info[index].flags &= ~DC_INCORE;
+  if (disk_cache_info[index].ref_count == 0 &&
+      !(disk_cache_info[index].flags & DC_DONT_REUSE))
+    disk_cache_info_free_push (&disk_cache_info[index]);
   pthread_mutex_unlock (&disk_cache_lock);
 }
 
@@ -636,7 +642,7 @@ pager_unlock_page (struct user_pager_info *pager, vm_offset_t page)
       if (err == ENOSPC)
 	ext2_warning ("This filesystem is out of space, and will now crash.  Bye!");
       else if (err)
-	ext2_warning ("inode=%Ld, page=0x%zx: %s",
+	ext2_warning ("inode=%Ld, page=0x%lx: %s",
 		      node->cache_id, page, strerror (err));
 
       return err;
@@ -840,13 +846,50 @@ int disk_cache_blocks;
 hurd_ihash_t disk_cache_bptr;
 /* Cached blocks' info.  */
 struct disk_cache_info *disk_cache_info;
-/* Hint index for which cache block to reuse next.  */
-int disk_cache_hint;
 /* Lock for these structures.  */
 pthread_mutex_t disk_cache_lock;
 /* Fired when a re-association is done.  */
 pthread_cond_t disk_cache_reassociation;
 
+/* Linked list of potentially unused blocks. */
+static struct disk_cache_info *disk_cache_info_free;
+static pthread_mutex_t disk_cache_info_free_lock;
+
+/* Get a reusable entry.  Must be called with disk_cache_lock
+   held.  */
+static struct disk_cache_info *
+disk_cache_info_free_pop (void)
+{
+  struct disk_cache_info *p;
+
+  do
+    {
+      pthread_mutex_lock (&disk_cache_info_free_lock);
+      p = disk_cache_info_free;
+      if (p)
+	{
+	  disk_cache_info_free = p->next;
+	  p->next = NULL;
+	}
+      pthread_mutex_unlock (&disk_cache_info_free_lock);
+    }
+  while (p && (p->flags & DC_DONT_REUSE || p->ref_count > 0));
+  return p;
+}
+
+/* Add P to the list of potentially re-usable entries.  */
+static void
+disk_cache_info_free_push (struct disk_cache_info *p)
+{
+  pthread_mutex_lock (&disk_cache_info_free_lock);
+  if (! p->next)
+    {
+      p->next = disk_cache_info_free;
+      disk_cache_info_free = p;
+    }
+  pthread_mutex_unlock (&disk_cache_info_free_lock);
+}
+
 /* Finish mapping initialization. */
 static void
 disk_cache_init (void)
@@ -857,6 +900,7 @@ disk_cache_init (void)
 
   pthread_mutex_init (&disk_cache_lock, NULL);
   pthread_cond_init (&disk_cache_reassociation, NULL);
+  pthread_mutex_init (&disk_cache_info_free_lock, NULL);
 
   /* Allocate space for block num -> in-memory pointer mapping.  */
   if (hurd_ihash_create (&disk_cache_bptr, HURD_IHASH_NO_LOCP))
@@ -867,19 +911,22 @@ disk_cache_init (void)
   if (!disk_cache_info)
     ext2_panic ("Cannot allocate space for disk cache info");
 
-  /* Initialize disk_cache_info.  */
-  for (int i = 0; i < disk_cache_blocks; i++)
+  /* Initialize disk_cache_info.  Start with the last entry so that
+     the first ends up at the front of the free list.  This keeps the
+     assertions at the end of this function happy.  */
+  for (int i = disk_cache_blocks - 1; i >= 0; i--)
     {
       disk_cache_info[i].block = DC_NO_BLOCK;
       disk_cache_info[i].flags = 0;
       disk_cache_info[i].ref_count = 0;
-#ifndef NDEBUG
+      disk_cache_info[i].next = NULL;
+      disk_cache_info_free_push (&disk_cache_info[i]);
+#ifdef DEBUG_DISK_CACHE
       disk_cache_info[i].last_read = DC_NO_BLOCK;
       disk_cache_info[i].last_read_xor
 	= DC_NO_BLOCK ^ DISK_CACHE_LAST_READ_XOR;
 #endif
     }
-  disk_cache_hint = 0;
 
   /* Map the superblock and the block group descriptors.  */
   block_t fixed_first = boffs_block (SBLOCK_OFFS);
@@ -947,7 +994,7 @@ disk_cache_return_unused (void)
 		       1);
   else
     {
-      printf ("ext2fs: disk cache is starving\n");
+      ext2_debug ("ext2fs: disk cache is starving\n");
 
       /* Give it some time.  This should happen rarely.  */
       sleep (1);
@@ -958,8 +1005,10 @@ disk_cache_return_unused (void)
 void *
 disk_cache_block_ref (block_t block)
 {
+  struct disk_cache_info *info;
   int index;
   void *bptr;
+  hurd_ihash_locp_t slot;
 
   assert (block < store->size >> log2_block_size);
 
@@ -968,7 +1017,7 @@ disk_cache_block_ref (block_t block)
 retry_ref:
   pthread_mutex_lock (&disk_cache_lock);
 
-  bptr = hurd_ihash_find (disk_cache_bptr, block);
+  bptr = hurd_ihash_locp_find (disk_cache_bptr, block, &slot);
   if (bptr)
     /* Already mapped.  */
     {
@@ -1004,34 +1053,10 @@ retry_ref:
     }
 
   /* Search for a block that is not in core and is not referenced.  */
-  index = disk_cache_hint;
-  while ((disk_cache_info[index].flags & DC_DONT_REUSE)
-	 || (disk_cache_info[index].ref_count))
-    {
-      ext2_debug ("reject %u -> %d (ref_count = %hu, flags = %#hx)",
-		  disk_cache_info[index].block, index,
-		  disk_cache_info[index].ref_count,
-		  disk_cache_info[index].flags);
-
-      /* Just move to next block.  */
-      index++;
-      if (index >= disk_cache_blocks)
-	index -= disk_cache_blocks;
-
-      /* If we return to where we started, than there is no suitable
-	 block. */
-      if (index == disk_cache_hint)
-	break;
-    }
-
-  /* The next place in the disk cache becomes the current hint.  */
-  disk_cache_hint = index + 1;
-  if (disk_cache_hint >= disk_cache_blocks)
-    disk_cache_hint -= disk_cache_blocks;
+  info = disk_cache_info_free_pop ();
 
   /* Is suitable place found?  */
-  if ((disk_cache_info[index].flags & DC_DONT_REUSE)
-      || disk_cache_info[index].ref_count)
+  if (info == NULL)
     /* No place is found.  Try to release some blocks and try
        again.  */
     {
@@ -1045,6 +1070,7 @@ retry_ref:
     }
 
   /* Suitable place is found.  */
+  index = info - disk_cache_info;
 
   /* Calculate pointer to data.  */
   bptr = (char *)disk_cache + (index << log2_block_size);
@@ -1091,12 +1117,13 @@ retry_ref:
 #endif
 
   /* Re-associate.  */
+
+  /* New association.  */
+  if (hurd_ihash_locp_add (disk_cache_bptr, slot, block, bptr))
+    ext2_panic ("Couldn't hurd_ihash_locp_add new disk block");
   if (disk_cache_info[index].block != DC_NO_BLOCK)
     /* Remove old association.  */
     hurd_ihash_remove (disk_cache_bptr, disk_cache_info[index].block);
-  /* New association.  */
-  if (hurd_ihash_add (disk_cache_bptr, block, bptr))
-    ext2_panic ("Couldn't hurd_ihash_add new disk block");
   assert (! (disk_cache_info[index].flags & DC_DONT_REUSE & ~DC_UNTOUCHED));
   disk_cache_info[index].block = block;
   assert (! disk_cache_info[index].ref_count);
@@ -1175,6 +1202,9 @@ disk_cache_block_deref (void *ptr)
   assert (! (disk_cache_info[index].flags & DC_UNTOUCHED));
   assert (disk_cache_info[index].ref_count >= 1);
   disk_cache_info[index].ref_count--;
+  if (disk_cache_info[index].ref_count == 0 &&
+      !(disk_cache_info[index].flags & DC_DONT_REUSE))
+    disk_cache_info_free_push (&disk_cache_info[index]);
   pthread_mutex_unlock (&disk_cache_lock);
 }
 
@@ -1200,8 +1230,6 @@ disk_cache_block_is_ref (block_t block)
 void
 create_disk_pager (void)
 {
-  pthread_t thread;
-  pthread_attr_t attr;
   error_t err;
 
   /* The disk pager.  */
