@@ -1,6 +1,6 @@
 /* Unparse run-time options
 
-   Copyright (C) 1995, 1996 Free Software Foundation, Inc.
+   Copyright (C) 1995, 1996, 1998 Free Software Foundation, Inc.
 
    This file is part of the GNU Hurd.
 
@@ -44,17 +44,17 @@ netfs_S_file_get_fs_options (struct protid *user,
   if (! err)
     {
 #if NOT_YET
-      rwlock_reader_lock (&netfs_fsys_lock);
+      pthread_rwlock_rdlock (&netfs_fsys_lock);
 #endif
       err = netfs_append_args (&argz, &argz_len);
 #if NOT_YET
-      rwlock_reader_unlock (&netfs_fsys_lock);
+      pthread_rwlock_unlock (&netfs_fsys_lock);
 #endif
     }
 
   if (! err)
     /* Move ARGZ from a malloced buffer into a vm_alloced one.  */
-    err = fshelp_return_malloced_buffer (argz, argz_len, data, data_len);
+    err = iohelp_return_malloced_buffer (argz, argz_len, data, data_len);
   else
     free (argz);
 
