@@ -1,5 +1,5 @@
-/* 
-   Copyright (C) 1994, 1996 Free Software Foundation
+/*
+   Copyright (C) 1994,96,2002 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
@@ -16,7 +16,7 @@
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */
 
 #include "priv.h"
-#include "fs_S.h"
+#include "trivfs_fs_S.h"
 #include <string.h>
 #include <unistd.h>
 
@@ -28,9 +28,9 @@ trivfs_S_file_statfs (struct trivfs_protid *cred,
   if (!trivfs_fsid)
     trivfs_fsid = getpid();
 
-  bzero (stb, sizeof (struct statfs));
+  memset (stb, 0, sizeof(struct statfs));
   stb->f_type = trivfs_fstype;
   stb->f_fsid = trivfs_fsid;
 
   return 0;
-}    
+}
