@@ -1267,7 +1267,10 @@ netfs_attempt_unlink (struct iouser *cred, struct node *dir,
      one we just got; if so, we must give this file another link
      so that when we delete the one we are asked for it doesn't go
      away entirely. */
-  if (np->references > 1)
+  struct references result;
+  refcounts_references (&np->refcounts, &result);
+
+  if (result.hard > 1)
     {
       char *newname = 0;
       int n = 0;
@@ -1819,7 +1822,7 @@ netfs_get_dirents (struct iouser *cred, struct node *np,
 		   mach_msg_type_number_t *datacnt,
 		   vm_size_t bufsiz, int *amt)
 {
-  void *buf;
+  void *buf = NULL;
   size_t our_bufsiz, allocsize;
   void *bp;
   char *userdp;
