@@ -1,6 +1,7 @@
 /* Set a file's translator.
 
-   Copyright (C) 1995,96,97,98,2001,02 Free Software Foundation, Inc.
+   Copyright (C) 1995,96,97,98,2001,02,13,14
+     Free Software Foundation, Inc.
    Written by Miles Bader <miles@gnu.org>
 
    This program is free software; you can redistribute it and/or
@@ -17,6 +18,7 @@
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */
 
+#include <assert.h>
 #include <hurd.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -25,6 +27,7 @@
 #include <error.h>
 #include <fcntl.h>
 #include <unistd.h>
+#include <sys/wait.h>
 
 #include <error.h>
 #include <argz.h>
@@ -43,12 +46,17 @@ const char *argp_program_version = STANDARD_HURD_VERSION (settrans);
 #define _STRINGIFY(arg) #arg
 #define STRINGIFY(arg) _STRINGIFY (arg)
 
+#define OPT_CHROOT_CHDIR	-1
+
 static struct argp_option options[] =
 {
   {"active",      'a', 0, 0, "Start TRANSLATOR and set it as NODE's active translator" },
+  {"start",       's', 0, 0, "Start the translator specified by the NODE's passive translator record and set it as NODE's active translator" },
   {"passive",     'p', 0, 0, "Change NODE's passive translator record (default)" },
   {"create",      'c', 0, 0, "Create NODE if it doesn't exist" },
   {"dereference", 'L', 0, 0, "If a translator exists, put the new one on top"},
+  {"pid-file",    'F', "FILENAME", 0, "When starting an active translator,"
+     " write its pid to this file"},
   {"pause",       'P', 0, 0, "When starting an active translator, prompt and"
      " wait for a newline on stdin before completing the startup handshake"},
   {"timeout",     't',"SEC",0, "Timeout for translator startup, in seconds"
@@ -56,10 +64,15 @@ static struct argp_option options[] =
   {"exclusive",   'x', 0, 0, "Only set the translator if there is not one already"},
   {"orphan",      'o', 0, 0, "Disconnect old translator from the filesystem "
 			     "(do not ask it to go away)"},
+  {"underlying",  'U', "NODE", 0, "Open NODE and hand it to the translator "
+				  "as the underlying node"},
 
   {"chroot",      'C', 0, 0,
    "Instead of setting the node's translator, take following arguments up to"
    " `--' and run that command chroot'd to the translated node."},
+  {"chroot-chdir",      OPT_CHROOT_CHDIR, "DIR", 0,
+   "Change to DIR before running the chrooted command.  "
+   "DIR must be an absolute path."},
 
   {0,0,0,0, "When setting the passive translator, if there's an active translator:"},
   {"goaway",      'g', 0, 0, "Ask the active translator to go away"},
@@ -77,6 +90,44 @@ static char *args_doc = "NODE [TRANSLATOR ARG...]";
 static char *doc = "Set the passive/active translator on NODE."
 "\vBy default the passive translator is set.";
 
+/* Authentication of the current process.  */
+uid_t *uids;
+gid_t *gids;
+size_t uids_len, gids_len;
+
+/* Initialize and populate the uids and gids vectors.  */
+error_t
+get_credentials (void)
+{
+  /* Fetch uids...  */
+  uids_len = geteuids (0, 0);
+  if (uids_len < 0)
+    return errno;
+
+  uids = malloc (uids_len * sizeof (uid_t));
+  if (! uids)
+    return ENOMEM;
+
+  uids_len = geteuids (uids_len, uids);
+  if (uids_len < 0)
+    return errno;
+
+  /* ... and gids.  */
+  gids_len = getgroups (0, 0);
+  if (gids_len < 0)
+    return errno;
+
+  gids = malloc (gids_len * sizeof (gid_t));
+  if (! uids)
+    return ENOMEM;
+
+  gids_len = getgroups (gids_len, gids);
+  if (gids_len < 0)
+    return errno;
+
+  return 0;
+}
+
 /* ---------------------------------------------------------------- */
 
 int
@@ -104,9 +155,13 @@ main(int argc, char *argv[])
   /* Various option flags.  */
   int passive = 0, active = 0, keep_active = 0, pause = 0, kill_active = 0,
       orphan = 0;
+  int start = 0;
+  char *pid_file = NULL;
   int excl = 0;
   int timeout = DEFAULT_TIMEOUT * 1000; /* ms */
+  char *underlying_node_name = NULL;
   char **chroot_command = 0;
+  char *chroot_chdir = "/";
 
   /* Parse our options...  */
   error_t parse_opt (int key, char *arg, struct argp_state *state)
@@ -118,6 +173,9 @@ main(int argc, char *argv[])
 	    node_name = arg;
 	  else			/* command */
 	    {
+	      if (start)
+		argp_error (state, "both --start and TRANSLATOR given");
+
 	      error_t err =
 		argz_create (state->argv + state->next - 1, &argz, &argz_len);
 	      if (err)
@@ -131,12 +189,27 @@ main(int argc, char *argv[])
 	  return EINVAL;
 
 	case 'a': active = 1; break;
+	case 's':
+	  start = 1;
+	  active = 1;	/* start implies active */
+	  break;
 	case 'p': passive = 1; break;
 	case 'k': keep_active = 1; break;
 	case 'g': kill_active = 1; break;
 	case 'x': excl = 1; break;
 	case 'P': pause = 1; break;
+	case 'F':
+	  pid_file = strdup (arg);
+	  if (pid_file == NULL)
+	    error(3, ENOMEM, "Failed to duplicate argument");
+	  break;
+
 	case 'o': orphan = 1; break;
+	case 'U':
+	  underlying_node_name = strdup (arg);
+	  if (underlying_node_name == NULL)
+	    error(3, ENOMEM, "Failed to duplicate argument");
+	  break;
 
 	case 'C':
 	  if (chroot_command)
@@ -163,6 +236,12 @@ main(int argc, char *argv[])
 	  argp_error (state, "--chroot command must be terminated with `--'");
 	  return EINVAL;
 
+	case OPT_CHROOT_CHDIR:
+	  if (arg[0] != '/')
+	    argp_error (state, "--chroot-chdir must be absolute");
+	  chroot_chdir = arg;
+	  break;
+
 	case 'c': lookup_flags |= O_CREAT; break;
 	case 'L': lookup_flags &= ~O_NOTRANS; break;
 
@@ -202,6 +281,26 @@ main(int argc, char *argv[])
 	active_flags = FS_TRANS_SET | FS_TRANS_EXCL;
     }
 
+  if (start)
+    {
+      /* Retrieve the passive translator record in argz.  */
+      mach_port_t node = file_name_lookup (node_name, lookup_flags, 0);
+      if (node == MACH_PORT_NULL)
+	error (4, errno, "%s", node_name);
+
+      char buf[1024];
+      argz = buf;
+      argz_len = sizeof (buf);
+
+      err = file_get_translator (node, &argz, &argz_len);
+      if (err == EINVAL)
+	error (4, 0, "%s: no passive translator record found", node_name);
+      if (err)
+	error (4, err, "%s", node_name);
+
+      mach_port_deallocate (mach_task_self (), node);
+    }
+
   if ((active || chroot_command) && argz_len > 0)
     {
       /* Error during file lookup; we use this to avoid duplicating error
@@ -221,6 +320,17 @@ main(int argc, char *argv[])
 	      getchar ();
 	    }
 
+	  if (pid_file != NULL)
+	    {
+	      FILE *h;
+	      h = fopen (pid_file, "w");
+	      if (h == NULL)
+		error (4, errno, "Failed to open pid file");
+
+	      fprintf (h, "%i\n", task2pid (task));
+	      fclose (h);
+	    }
+
 	  node = file_name_lookup (node_name, flags | lookup_flags, 0666);
 	  if (node == MACH_PORT_NULL)
 	    {
@@ -228,7 +338,20 @@ main(int argc, char *argv[])
 	      return open_err;
 	    }
 
-	  *underlying = node;
+	  if (underlying_node_name)
+	    {
+	      *underlying = file_name_lookup (underlying_node_name,
+					      flags | lookup_flags, 0666);
+	      if (! MACH_PORT_VALID (*underlying))
+		{
+		  /* For the error message.  */
+		  node_name = underlying_node_name;
+		  open_err = errno;
+		  return open_err;
+		}
+	    }
+	  else
+	    *underlying = node;
 	  *underlying_type = MACH_MSG_TYPE_COPY_SEND;
 
 	  return 0;
@@ -259,33 +382,79 @@ main(int argc, char *argv[])
 
   if (chroot_command)
     {
-      /* We will act as the parent filesystem would for a lookup
-	 of the active translator's root node, then use this port
-	 as our root directory while we exec the command.  */
+      pid_t child;
+      int status;
+      switch ((child = fork ()))
+	{
+	case -1:
+	  error (6, errno, "fork");
 
-      char retry_name[1024];	/* XXX */
-      retry_type do_retry;
-      mach_port_t root;
-      err = fsys_getroot (active_control,
-			  MACH_PORT_NULL, MACH_MSG_TYPE_COPY_SEND,
-			  NULL, 0, NULL, 0, 0, &do_retry, retry_name, &root);
-      mach_port_deallocate (mach_task_self (), active_control);
-      if (err)
-	error (6, err, "fsys_getroot");
-      err = hurd_file_name_lookup_retry (&_hurd_ports_use, &getdport, 0,
-					 do_retry, retry_name, 0, 0,
-					 &root);
-      if (err)
-	error (6, err, "cannot resolve root port");
+	case 0:; /* Child.  */
+	  /* We will act as the parent filesystem would for a lookup
+	     of the active translator's root node, then use this port
+	     as our root directory while we exec the command.  */
 
-      if (setcrdir (root))
-	error (7, errno, "cannot install root port");
-      mach_port_deallocate (mach_task_self (), root);
-      if (chdir ("/"))
-	error (8, errno, "cannot chdir to new root");
+	  char retry_name[1024];	/* XXX */
+	  retry_type do_retry;
+	  mach_port_t root;
+	  file_t executable;
+	  char *prefixed_name;
 
-      execvp (chroot_command[0], chroot_command);
-      error (8, errno, "cannot execute %s", chroot_command[0]);
+	  err = get_credentials ();
+	  if (err)
+	    error (6, err, "getting credentials");
+
+	  err = fsys_getroot (active_control,
+			      MACH_PORT_NULL, MACH_MSG_TYPE_COPY_SEND,
+			      uids, uids_len, gids, gids_len, 0,
+			      &do_retry, retry_name, &root);
+	  mach_port_deallocate (mach_task_self (), active_control);
+	  if (err)
+	    error (6, err, "fsys_getroot");
+	  err = hurd_file_name_lookup_retry (&_hurd_ports_use, &getdport, 0,
+					     do_retry, retry_name, 0, 0,
+					     &root);
+	  if (err)
+	    error (6, err, "cannot resolve root port");
+
+	  if (setcrdir (root))
+	    error (7, errno, "cannot install root port");
+	  mach_port_deallocate (mach_task_self (), root);
+	  if (chdir (chroot_chdir))
+	    error (8, errno, "%s", chroot_chdir);
+
+	  /* Lookup executable in PATH.  */
+	  executable = file_name_path_lookup (chroot_command[0],
+					      getenv ("PATH"),
+					      O_EXEC, 0,
+					      &prefixed_name);
+	  if (MACH_PORT_VALID (executable))
+	    {
+	      err = mach_port_deallocate (mach_task_self (), executable);
+	      assert_perror (err);
+	      if (prefixed_name)
+		chroot_command[0] = prefixed_name;
+	    }
+
+	  execvp (chroot_command[0], chroot_command);
+	  error (8, errno, "cannot execute %s", chroot_command[0]);
+	  break;
+
+	default: /* Parent.  */
+	  if (waitpid (child, &status, 0) != child)
+	    error (8, errno, "waitpid on %d", child);
+
+	  err = fsys_goaway (active_control, goaway_flags);
+	  if (err && err != EBUSY)
+	    error (9, err, "fsys_goaway");
+
+	  if (WIFSIGNALED (status))
+	    error (WTERMSIG (status) + 128, 0,
+		   "%s for child %d", strsignal (WTERMSIG (status)), child);
+	  if (WEXITSTATUS (status) != 0)
+	    error (WEXITSTATUS (status), 0,
+		   "Error %d for child %d", WEXITSTATUS (status), child);
+	}
     }
 
   return 0;
